/*----------------------------------------------------------------------
  File    : match.c
  Contents: compute optimum cost row to column matching
  Author  : Christian Borgelt
  History : 2008.09.21 file created
            2010.10.11 adapted to modified module tabread
            2010.12.11 adapted to a generic error reporting function
            2010.12.20 condensed output added (pure row permutation)
            2011.02.04 extended to maximum matchings (negate matrix)
            2013.08.09 adapted to higher compiler warning level
            2013.08.13 adapted to definition of type DIMID
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include "escape.h"
#include "tabread.h"
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "match"
#define DESCRIPTION "compute optimum cost row to column matching"
#define VERSION     "version 2.1 (2014.10.24)         " \
                    "(c) 2008-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes   0 to  -4 defined in matrix.h */
#define E_STDIN     (-5)        /* double assignment of stdin */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_COLCNT    (-9)        /* too many rows for columns */
/* error codes  -16 to -18 defined in matrix.h */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_COLCNT   -9 */  "need at least as many columns as rows",
  /*    -10 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_VALUE   -16 */  "%s:%d(%d): invalid value '%s'",
  /* E_FLDCNT  -17 */  "%s:%d(%d): wrong number of fields/columns",
  /* E_RECCNT  -18 */  "%s:%d(%d): wrong number of records/rows",
  /*           -19 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static TABREAD *tread = NULL;   /* table reader */
static MATRIX  *mat   = NULL;   /* data matrix */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (mat)   mat_delete(mat); \
  if (tread) trd_delete(tread, 1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of file to read */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  char    seps[4]  = " \n";     /* separators for output */
  int     cmax     = 0;         /* flag for maximum cost matching */
  int     cond     = 0;         /* flag for condensed output */
  double  cost     = 0;         /* cost of row to column matching */
  DIMID   *map;                 /* row to column mapping */
  DIMID   r, c;                 /* number of rows and columns */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] infile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-c       condensed output "
                    "(print only a vector of numbers)\n");
    printf("-i       find maximum cost matching "
                    "(default: minimum)\n");
    printf("-r#      record  separators         "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators         "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters         "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters         "
                    "(default: \"#\")\n");
    printf("infile   matrix file to read (numbers only)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse the arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'i': cmax   = 1;            break;
          case 'c': cond   = 1;            break;
          case 'r': optarg = &recseps;     break;
          case 'f': optarg = &fldseps;     break;
          case 'b': optarg = &blanks;      break;
          case 'C': optarg = &comment;     break;
          default : error(E_OPTION, *--s); break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_inp = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- read matrix file --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_inp) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  k = mat_readx(&mat, tread, -1, -1);    /* read the matrix */
  if (k) error(k, TRD_INFO(tread));
  trd_delete(tread, 1);         /* close the input file and */
  tread = NULL;                 /* delete the table reader */
  r = mat_rowcnt(mat);          /* get the number of rows */
  c = mat_colcnt(mat);          /* and the number of columns */
  fprintf(stderr, "[%"DIMID_FMT" row(s),",    r);
  fprintf(stderr, " %"DIMID_FMT" column(s)]", c);
  if (c < r) error(E_COLCNT);   /* check for more columns than rows */
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- find matching --- */
  t = clock();                  /* start the timer */
  fprintf(stderr, "computing best matching ... ");
  if (cmax)                     /* if to find maximum cost matching, */
    mat_negate(mat, 0);         /* negate the cost matrix */
  map = (DIMID*)malloc((size_t)i *sizeof(DIMID));
  if (!map) error(E_NOMEM);     /* create a row to column map and */
  cost = mat_match(mat, map);   /* compute row to column matching */
  if (cmax) cost = -cost;       /* adapt cost for max. matching */
  fprintf(stderr, "[cost: %g] done [%.2fs].\n", cost, SEC_SINCE(t));

  /* --- print found matching --- */
  if (cond) {                   /* if condensed output */
    if (fldseps) seps[0] = (char)esc_decode(fldseps, NULL);
    if (recseps) seps[1] = (char)esc_decode(recseps, NULL);
    for (c = 0; c < r; c++) {   /* traverse the matrix rows */
      if (c > 0) putchar(seps[0]);     /* print a separator */
      printf("%"DIMID_FMT, map[c]);
    }                           /* print the column mapped to */
    putchar(seps[1]); }         /* terminate the output record */
  else {                        /* if normal/verbose output */
    printf("row to column map:\n");
    for (c = 0; c < r; c++)     /* print the row to column map */
      printf("%"DIMID_FMT" -> %"DIMID_FMT"\n", c, map[c]);
    printf("cost: %g\n", cost); /* print the total matching cost */
  }                             /* (sum of selected matrix elements) */
  free(map);                    /* delete the row to column map */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
